import grpc
import json
import os
import pytest
import random
import tempfile
from concurrent import futures
from unittest.mock import patch

import amaas.grpc.aio
from .mock_server import MockScanServicer
from amaas.grpc.exception import AMaasErrorCode
from amaas.grpc.exception import AMaasException


NUM_DATA_LOOP = 128
_, TEST_DATA_FILE_NAME = tempfile.mkstemp()
SERVER_PORT = random.randint(49152, 65535)


#
# Test init_by_region method
#
@patch("amaas.grpc.aio._init_by_region_util")
def test_init_by_region(utilmock):
    amaas.grpc.aio.init_by_region("us-east-1", "dummy_key")

    # ensure True is passed to is_aio_channel parameter
    utilmock.assert_called_with("us-east-1", "dummy_key", True, None, True)


#
# Test init method
#
@patch("amaas.grpc.aio._init_util")
def test_init(utilmock):
    amaas.grpc.aio.init("us-east-1", "dummy_key")

    # ensure False is passed to is_aio_channel parameter
    utilmock.assert_called_with("us-east-1", "dummy_key", False, None, True)


#
# Create a mock gRPC server with MockScanServicer
#
@pytest.fixture(scope="module", autouse=True)
def run_setup():
    server = grpc.server(futures.ThreadPoolExecutor(max_workers=10))

    amaas.grpc.protos.scan_pb2_grpc.add_ScanServicer_to_server(
        MockScanServicer(), server
    )
    server.add_insecure_port(f"[::]:{SERVER_PORT}")
    server.start()
    yield server
    server.stop(None)


#
# This routine is to create a data file for testing.
# The binary file contains of NUM_DATA_LOOP of consecutive bytes from 0x00 to 0xff.
# This is to make checking the correctness of the content of the retrieval easier.
# The grpc client receives a sequence of instructions from grpc server to retrieve chunks of data
# from a data file during scanning.
# One of the important checking is to ensure the data retrived is what is actually rqeuested by the server.
#
@pytest.fixture(scope="session", autouse=True)
def create_data_file():
    with open(TEST_DATA_FILE_NAME, "wb") as binary_file:
        for j in range(NUM_DATA_LOOP):
            for i in range(256):
                # Convert the number to a single byte and write to the file
                byte_value = i.to_bytes(1, byteorder="big")
                binary_file.write(byte_value)
    yield
    os.remove(TEST_DATA_FILE_NAME)


#
# Testing the SDK scan_file method failed with MSG_ID_ERR_FILE_NOT_FOUND exception.
#
@pytest.mark.asyncio
async def test_scan_file_not_found():
    handle = grpc.aio.insecure_channel(f"localhost:{SERVER_PORT}")
    NOT_EXIST_FILE = "012345.txt"
    with pytest.raises(AMaasException) as exc_info:
        await amaas.grpc.scan_file(handle, NOT_EXIST_FILE)
    assert exc_info.value.args[0] == AMaasErrorCode.MSG_ID_ERR_FILE_NOT_FOUND
    assert exc_info.value.args[1] == NOT_EXIST_FILE


#
# Testing the SDK scan_file method failed with MSG_ID_ERR_FILE_NO_PERMISSION exception
#
@pytest.mark.asyncio
async def test_scan_file_no_permission():
    handle = grpc.insecure_channel(f"localhost:{SERVER_PORT}")
    NOT_PERMISSION_FILE = tempfile.NamedTemporaryFile().name
    open(NOT_PERMISSION_FILE, "x")
    os.chmod(NOT_PERMISSION_FILE, 0x000)
    with pytest.raises(AMaasException) as exc_info:
        await amaas.grpc.scan_file(handle, NOT_PERMISSION_FILE)
    assert exc_info.value.args[0] == AMaasErrorCode.MSG_ID_ERR_FILE_NO_PERMISSION
    assert exc_info.value.args[1] == NOT_PERMISSION_FILE
    os.remove(NOT_PERMISSION_FILE)


#
# Testing the SDK scan_buffer method succeeded without viruses.
#
@pytest.mark.asyncio
async def test_scan_buffer_success():
    handle = grpc.aio.insecure_channel(f"localhost:{SERVER_PORT}")
    with open(TEST_DATA_FILE_NAME, mode="rb") as bfile:
        buffer = bfile.read()
        response = await amaas.grpc.aio.scan_buffer(handle, buffer, TEST_DATA_FILE_NAME)
        assert json.loads(response)["scanResult"] == 0


#
# Testing the SDK scan_buffer method succeeded with viruses.
#
@pytest.mark.asyncio
async def test_scan_buffer_virus():
    handle = grpc.aio.insecure_channel(f"localhost:{SERVER_PORT}")
    with open(TEST_DATA_FILE_NAME, mode="rb") as bfile:
        buffer = bfile.read()
        response = await amaas.grpc.aio.scan_buffer(handle, buffer, "virus")
        jobj = json.loads(response)
        assert jobj["scanResult"] == 1
        assert jobj["foundMalwares"] == ["virus1", "virus2"]


#
# Test different exceptions throwed by scan_buffer API.
#
@pytest.mark.asyncio
@pytest.mark.parametrize(
    "error_type, expected_exception",
    [
        (
            MockScanServicer.IDENTIFIER_EXCEED_RATE,
            [AMaasErrorCode.MSG_ID_ERR_RATE_LIMIT_EXCEEDED],
        ),
        (
            MockScanServicer.IDENTIFIER_MISMATCHED,
            [
                AMaasErrorCode.MSG_ID_ERR_UNEXPECTED_CMD_AND_STAGE,
                amaas.grpc.scan_pb2.CMD_QUIT,
                amaas.grpc.scan_pb2.STAGE_RUN,
            ],
        ),
        (
            MockScanServicer.IDENTIFIER_UNKNOWN_CMD,
            [AMaasErrorCode.MSG_ID_ERR_UNKNOWN_CMD, MockScanServicer.UNKNOWN_CMD],
        ),
        (
            MockScanServicer.IDENTIFIER_GRPC_ERROR,
            [AMaasErrorCode.MSG_ID_GRPC_ERROR, grpc.StatusCode.INTERNAL.value[0]],
        ),
    ],
)
async def test_scan_buffer_exceptions(error_type, expected_exception):
    handle = grpc.aio.insecure_channel(f"localhost:{SERVER_PORT}")
    with open(TEST_DATA_FILE_NAME, mode="rb") as bfile:
        buffer = bfile.read()
        with pytest.raises(AMaasException) as exc_info:
            await amaas.grpc.aio.scan_buffer(handle, buffer, error_type)
        for cnt in range(len(expected_exception)):
            assert exc_info.value.args[cnt] == expected_exception[cnt]



import os
import grpc
from unittest.mock import patch

import amaas.grpc.util


#
# Test _init_by_region_util is invoked with correct arguments.
#
@patch("amaas.grpc.util._init_util")
def test_init_by_region_util(utilmock):
    amaas.grpc.util._init_by_region_util("ap-southeast-2", "dummy_key")

    utilmock.assert_called_with(
        "antimalware.au-1.cloudone.trendmicro.com:443", "dummy_key", True, None, False
    )


#
# Test insecure channel is created.
#
def test_insecure_channel():
    channel = amaas.grpc.util._init_by_region_util(
        "us-east-1", None, False, None, False
    )
    assert type(channel) is grpc._channel.Channel


#
# Test insecure aio channel is created.
#
def test_aio_insecure_channel():
    channel = amaas.grpc.util._init_by_region_util(
        "us-east-1", None, is_aio_channel=True
    )
    assert type(channel) is grpc.aio._channel.Channel


#
# Test secure channel is created.
#
def test_secure_channel():
    channel = amaas.grpc.util._init_by_region_util("us-east-1", None, True, None, False)
    assert type(channel) is grpc._channel.Channel


#
# Test secure aio channel is created.
#
def test_aio_secure_channel():
    channel = amaas.grpc.util._init_by_region_util(
        "us-east-1", None, True, None, is_aio_channel=True
    )
    assert type(channel) is grpc.aio._channel.Channel


#
# Test default SSL only channel.
#
@patch("grpc.secure_channel")
def test_def_ssl_only_channel(channel_mock):
    amaas.grpc.util._init_by_region_util("us-east-1", None, True, None, False)

    args = channel_mock.call_args.args
    assert type(args[1]._credentials) is grpc._cython.cygrpc.SSLChannelCredentials


#
# Test SSL only channel.
#
# @patch("builtins.open", new_callable=mock_open, read_data=SERVER_CERT)
@patch("grpc.secure_channel")
def test_ssl_only_channel(channel_mock):
    dir_path = os.path.dirname(os.path.realpath(__file__))
    amaas.grpc.util._init_by_region_util(
        "us-east-1", None, True, dir_path + "/fake_server_cert.pem", False
    )

    args = channel_mock.call_args.args
    assert type(args[1]._credentials) is grpc._cython.cygrpc.SSLChannelCredentials


#
# Test api key composite channel credential.
#
@patch("amaas.grpc.util._GrpcAuth")
@patch("grpc.secure_channel")
def test_composite_channel_with_apikey(channel_mock, auth_mock):
    dir_path = os.path.dirname(os.path.realpath(__file__))
    amaas.grpc.util._init_by_region_util(
        "us-east-1", "abcabc12345678", True, dir_path + "/fake_server_cert.pem", False
    )

    auth_mock.assert_called_with(
        "ApiKey abcabc12345678",
    )

    args = channel_mock.call_args.args
    assert type(args[1]._credentials) is grpc._cython.cygrpc.CompositeChannelCredentials



from setuptools import setup, find_namespace_packages

with open("README.md", "r") as fh:
    long_description = fh.read()
with open("VERSION", "r") as fh:
    package_version = fh.read().strip()
setup(install_requires=['grpcio==1.60.0', 'protobuf==4.25.1'],
    name="visionone-filesecurity",
    version=package_version,
    author="Trend Micro VisionOne File Security Team",
    description="Trend Micro VisionOne File Security SDK for python",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/trendmicro/tm-v1-fs-python-sdk",
    packages=find_namespace_packages(exclude=["tests*", "examples"]),
    package_data={"amaas": ["grpc/protos/*"]},
    include_package_data=True,
    classifiers=[
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "License :: OSI Approved :: MIT License",
    ],
    python_requires=">=3.7",
)


import threading
import logging
from typing import BinaryIO, List

import grpc
import os
import io

from .protos import scan_pb2
from .protos import scan_pb2_grpc
from .exception import AMaasException
from .exception import AMaasErrorCode
from .util import _init_by_region_util
from .util import _init_util
from .util import _validate_tags
from .util import _digest_hex
from .util import APP_NAME_HEADER, APP_NAME_FILE_SCAN

logger = logging.getLogger(__name__)
logger.addHandler(logging.StreamHandler())
LOG_LEVEL = os.environ.get('LOG_LEVEL', 'INFO')
logger.setLevel(LOG_LEVEL)
logger.propagate = False

timeout_in_seconds = int(os.environ.get('TM_AM_SCAN_TIMEOUT_SECS', 300))


class _Pipeline:
    """
    Class to allow a single element pipeline between producer and consumer.
    """

    def __init__(self):
        self._message = 0
        self._producer_lock = threading.Lock()
        self._consumer_lock = threading.Lock()
        self._consumer_lock.acquire()

    def get_message(self):
        self._consumer_lock.acquire()
        message = self._message
        self._producer_lock.release()
        return message

    def set_message(self, message):
        self._producer_lock.acquire()
        self._message = message
        self._consumer_lock.release()


def init_by_region(region, api_key, enable_tls=True, ca_cert=None):
    return _init_by_region_util(region, api_key, enable_tls, ca_cert, False)


def init(host, api_key=None, enable_tls=False, ca_cert=None):
    return _init_util(host, api_key, enable_tls, ca_cert, False)


def _generate_messages(pipeline: _Pipeline, data_reader: BinaryIO, bulk: bool, stats: dict) -> None:
    responses = []

    while True:
        for r in responses:
            if r[0] == "INIT":
                response = r[1]
            elif r[0] == "RUN":
                offset = r[1]
                length = r[2]
                data_reader.seek(offset)
                chunk = data_reader.read(length)
                response = scan_pb2.C2S(
                    stage=scan_pb2.STAGE_RUN,
                    file_name=None,
                    rs_size=0,
                    offset=offset,
                    chunk=chunk,
                )
                stats["total_upload"] = stats.get("total_upload", 0) + len(chunk)
            else:
                raise AMaasException(AMaasErrorCode.MSG_ID_ERR_UNEXPECTED_CMD_AND_STAGE, "None", r[0])
            yield response

        responses.clear()
        message = pipeline.get_message()

        if message.stage == scan_pb2.STAGE_INIT:
            logger.debug("stage INIT")
            responses.append(("INIT", message))
        elif message.stage == scan_pb2.STAGE_RUN:
            if message.cmd != scan_pb2.CMD_RETR:
                raise AMaasException(AMaasErrorCode.MSG_ID_ERR_UNEXPECTED_CMD_AND_STAGE, message.cmd, message.stage)

            length = []
            offset = []

            if bulk:
                offset = message.bulk_offset[:]
                length = message.bulk_length[:]

                if len(length) > 1:
                    logger.debug("bulk transfer triggered")
            else:
                offset.append(message.offset)
                length.append(message.length)

            for i in range(len(length)):
                logger.debug(f"stage RUN, try to read {length[i]} at offset {offset[i]}")
                responses.append(("RUN", offset[i], length[i]))
        elif message.stage == scan_pb2.STAGE_FINI:
            if message.cmd != scan_pb2.CMD_QUIT:
                raise AMaasException(AMaasErrorCode.MSG_ID_ERR_UNEXPECTED_CMD_AND_STAGE, message.cmd, message.stage)

            logger.debug("final stage, quit generating C2S messages...")
            break
        else:
            logger.debug("unknown stage.....!!!")
            raise AMaasException(AMaasErrorCode.MSG_ID_ERR_UNKNOWN_STAGE, message.stage)


def quit(handle):
    handle.close()


def _scan_data(channel: grpc.Channel, data_reader: BinaryIO, size: int, identifier: str, tags: List[str],
               pml: bool, feedback: bool, verbose: bool) -> str:
    _validate_tags(tags)
    stub = scan_pb2_grpc.ScanStub(channel)
    pipeline = _Pipeline()
    stats = {}
    result = None
    bulk = True

    try:
        metadata = (
            (APP_NAME_HEADER, APP_NAME_FILE_SCAN),
        )
        responses = stub.Run(_generate_messages(pipeline, data_reader, bulk, stats), timeout=timeout_in_seconds,
                             metadata=metadata)
        message = scan_pb2.C2S(stage=scan_pb2.STAGE_INIT,
                               file_name=identifier,
                               rs_size=size,
                               offset=0,
                               chunk=None,
                               trendx=pml,
                               tags=tags,
                               file_sha1="sha1:" + _digest_hex(data_reader, "sha1"),
                               file_sha256="sha256:" + _digest_hex(data_reader, "sha256"),
                               bulk=bulk,
                               spn_feedback=feedback,
                               verbose=verbose)

        pipeline.set_message(message)

        for response in responses:
            if response.cmd == scan_pb2.CMD_RETR:
                pipeline.set_message(response)
            elif response.cmd == scan_pb2.CMD_QUIT:
                result = response.result
                pipeline.set_message(response)
                logger.debug("receive QUIT, exit loop...")
                break
            else:
                logger.debug("unknown command...")
                raise AMaasException(AMaasErrorCode.MSG_ID_ERR_UNKNOWN_CMD, response.cmd)

        total_upload = stats.get("total_upload", 0)
        logger.debug(f"total upload {total_upload} bytes")

    except AMaasException:
        raise
    except grpc.RpcError as rpc_error:
        if "429" in str(rpc_error):
            raise AMaasException(AMaasErrorCode.MSG_ID_ERR_RATE_LIMIT_EXCEEDED)
        elif rpc_error.code() == grpc.StatusCode.UNAUTHENTICATED:
            raise AMaasException(AMaasErrorCode.MSG_ID_ERR_KEY_AUTH_FAILED)
        else:
            raise AMaasException(AMaasErrorCode.MSG_ID_GRPC_ERROR, rpc_error.code().value[0], rpc_error.details())
    except Exception as err:
        raise AMaasException(AMaasErrorCode.MSG_ID_ERR_UNEXPECTED_ERROR, str(err))

    return result


def scan_file(channel: grpc.Channel, file_name: str, tags: List[str] = None,
              pml: bool = False, feedback: bool = False, verbose: bool = False) -> str:
    try:
        f = open(file_name, "rb")
        fid = os.path.basename(file_name)
        n = os.stat(file_name).st_size
    except FileNotFoundError as err:
        logger.debug("File not exist: " + str(err))
        raise AMaasException(AMaasErrorCode.MSG_ID_ERR_FILE_NOT_FOUND, file_name)
    except (PermissionError, IOError) as err:
        logger.debug("Permission error: " + str(err))
        raise AMaasException(AMaasErrorCode.MSG_ID_ERR_FILE_NO_PERMISSION, file_name)

    return _scan_data(channel, f, n, fid, tags, pml, feedback, verbose)


def scan_buffer(channel: grpc.Channel, bytes_buffer: bytes, uid: str, tags: List[str] = None,
                pml: bool = False, feedback: bool = False, verbose: bool = False) -> str:
    f = io.BytesIO(bytes_buffer)
    return _scan_data(channel, f, len(bytes_buffer), uid, tags, pml, feedback, verbose)



import grpc
import json
import os
import pytest
import random
import tempfile
import uuid
from concurrent import futures
from unittest.mock import patch

import amaas.grpc
from .mock_server import MockScanServicer
from amaas.grpc.exception import AMaasErrorCode
from amaas.grpc.exception import AMaasException


NUM_DATA_LOOP = 128
_, TEST_DATA_FILE_NAME = tempfile.mkstemp()
SERVER_PORT = random.randint(49152, 65535)


@patch("amaas.grpc._init_by_region_util")
def test_init_by_region(utilmock):
    amaas.grpc.init_by_region("ap-southeast-2", "dummy_key")

    # ensure False is passed to is_aio_channel parameter
    utilmock.assert_called_with("ap-southeast-2", "dummy_key", True, None, False)


@patch("amaas.grpc._init_util")
def test_init(utilmock):
    amaas.grpc.init("ap-southeast-2", "dummy_key")

    # ensure False is passed to is_aio_channel parameter
    utilmock.assert_called_with("ap-southeast-2", "dummy_key", False, None, False)


#
# Create a mock gRPC server with MockScanServicer
#
@pytest.fixture(scope="module", autouse=True)
def run_setup():
    server = grpc.server(futures.ThreadPoolExecutor(max_workers=10))

    amaas.grpc.protos.scan_pb2_grpc.add_ScanServicer_to_server(
        MockScanServicer(), server
    )
    server.add_insecure_port(f"[::]:{SERVER_PORT}")
    server.start()
    yield server
    server.stop(None)


#
# This routine is to create a data file for testing.
# The binary file contains of NUM_DATA_LOOP of consecutive bytes from 0x00 to 0xff.
# This is to make checking the correctness of the content of the retrieval easier.
# The grpc client receives a sequence of instructions from grpc server to retrieve chunks of data
# from a data file during scanning.
# One of the important checking is to ensure the data retrived is what is actually rqeuested by the server.
#
@pytest.fixture(scope="session", autouse=True)
def create_data_file():
    with open(TEST_DATA_FILE_NAME, "wb") as binary_file:
        for j in range(NUM_DATA_LOOP):
            for i in range(256):
                # Convert the number to a single byte and write to the file
                byte_value = i.to_bytes(1, byteorder="big")
                binary_file.write(byte_value)
    yield
    os.remove(TEST_DATA_FILE_NAME)


#
# The method is to verify the chunk matching the data chunk at the given location in the file.
# @param buf buf of the data chunk
# @param offset location in the data file
# @return True if matched False otherwise.
#
def verify_buf_with_data(buf, offset):
    blen = len(buf)
    cnt = 0
    for posn in range(offset, blen):
        val = posn % 256
        # byte_value = val.to_bytes(1, byteorder='big')
        assert buf[cnt] == val
        cnt += 1


#
# Testing authentication failure.
#
def test_scan_file_failed_authentication():
    dir_path = os.path.dirname(os.path.realpath(__file__))
    handle = amaas.grpc.init_by_region("us-east-1", "api-key", True, None)
    with pytest.raises(AMaasException) as exc_info:
        amaas.grpc.scan_file(handle, dir_path + "/fake_server_cert.pem")
    assert exc_info.value.args[0] == AMaasErrorCode.MSG_ID_ERR_KEY_AUTH_FAILED


#
# Testing the generate_message method correctly processing the initial message from grpc client
#
def test_generate_message_init():
    pipeline = amaas.grpc._Pipeline()
    stats = {}
    with open(TEST_DATA_FILE_NAME, "rb") as f:
        size = os.stat(TEST_DATA_FILE_NAME).st_size
        server_resp = amaas.grpc.scan_pb2.C2S(
            stage=amaas.grpc.scan_pb2.STAGE_INIT,
            file_name=TEST_DATA_FILE_NAME,
            rs_size=size,
            offset=0,
            chunk=None,
            tags=None,
        )

        pipeline.set_message(server_resp)
        c2s_msg = next(amaas.grpc._generate_messages(pipeline, f, True, stats))
        assert c2s_msg.stage == amaas.grpc.scan_pb2.STAGE_INIT
        assert c2s_msg.offset == 0
        assert c2s_msg.rs_size == size
        assert c2s_msg.file_name == TEST_DATA_FILE_NAME


#
# Testing the generate_message method correctly processing a number of back and forth
# data retrieval message from grpc server.
#
def test_generate_message_data():
    pipeline = amaas.grpc._Pipeline()
    stats = {}
    f = open(TEST_DATA_FILE_NAME, "rb")
    size = os.stat(TEST_DATA_FILE_NAME).st_size
    #
    # lets loop 5 times to check
    #
    for _ in range(5):
        start = random.randint(0, size - 1)
        end = random.randint(start, size)
        server_resp = amaas.grpc.scan_pb2.S2C(
            stage=amaas.grpc.scan_pb2.STAGE_RUN,
            cmd=amaas.grpc.scan_pb2.CMD_RETR,
            bulk_offset=[start],
            bulk_length=[end - start],
        )

        pipeline.set_message(server_resp)
        c2s_msg = next(amaas.grpc._generate_messages(pipeline, f, True, stats))
        assert c2s_msg.stage == amaas.grpc.scan_pb2.STAGE_RUN
        assert c2s_msg.offset == start
        verify_buf_with_data(c2s_msg.chunk, start)


#
# Testing the generate_message method correctly raise exception when recieved
# wrong command in STAGE RUN from server.
#
def test_generate_message_data_wrong_cmd_exception():
    pipeline = amaas.grpc._Pipeline()
    stats = {}
    with open(TEST_DATA_FILE_NAME, "rb") as f:
        size = os.stat(TEST_DATA_FILE_NAME).st_size
        start = random.randint(0, size - 1)
        end = random.randint(start, size)
        server_resp = amaas.grpc.scan_pb2.S2C(
            stage=amaas.grpc.scan_pb2.STAGE_RUN,
            cmd=amaas.grpc.scan_pb2.CMD_QUIT,
            bulk_offset=[start],
            bulk_length=[end - start],
        )
        pipeline.set_message(server_resp)
        with pytest.raises(AMaasException) as exc_info:
            next(amaas.grpc._generate_messages(pipeline, f, True, stats))
        assert (
            exc_info.value.args[0] == AMaasErrorCode.MSG_ID_ERR_UNEXPECTED_CMD_AND_STAGE
        )
        assert exc_info.value.args[1] == amaas.grpc.scan_pb2.CMD_QUIT
        assert exc_info.value.args[2] == amaas.grpc.scan_pb2.STAGE_RUN


#
# Testing the generate_message method correctly processing the final message from grpc server.
#
def test_generate_message_final():
    pipeline = amaas.grpc._Pipeline()
    stats = {}
    f = open(TEST_DATA_FILE_NAME, "rb")
    server_resp = amaas.grpc.scan_pb2.S2C(
        stage=amaas.grpc.scan_pb2.STAGE_FINI, cmd=amaas.grpc.scan_pb2.CMD_QUIT
    )

    pipeline.set_message(server_resp)
    client_msg = next(amaas.grpc._generate_messages(pipeline, f, True, stats), None)
    assert client_msg is None


#
# Testing the generate_message method correctly raise exception when recieved wrong command in STAGE_FINI from server.
#
def test_generate_message_final_wrong_cmd_exception():
    pipeline = amaas.grpc._Pipeline()
    stats = {}
    with open(TEST_DATA_FILE_NAME, "rb") as f:
        size = os.stat(TEST_DATA_FILE_NAME).st_size
        start = random.randint(0, size - 1)
        end = random.randint(start, size)
        server_resp = amaas.grpc.scan_pb2.S2C(
            stage=amaas.grpc.scan_pb2.STAGE_FINI,
            cmd=amaas.grpc.scan_pb2.CMD_RETR,
            offset=start,
            length=end - start,
        )
        pipeline.set_message(server_resp)
        with pytest.raises(AMaasException) as exc_info:
            next(amaas.grpc._generate_messages(pipeline, f, True, stats))
        assert (
            exc_info.value.args[0] == AMaasErrorCode.MSG_ID_ERR_UNEXPECTED_CMD_AND_STAGE
        )
        assert exc_info.value.args[1] == amaas.grpc.scan_pb2.CMD_RETR
        assert exc_info.value.args[2] == amaas.grpc.scan_pb2.STAGE_FINI


#
# Testing the generate_message method correctly processing the unknown stage message from grpc server.
#
def test_generate_message_unknwon_stage():
    pipeline = amaas.grpc._Pipeline()
    stats = {}
    f = open(TEST_DATA_FILE_NAME, "rb")
    UNKNOWN_STAGE = 9999
    server_resp = amaas.grpc.scan_pb2.S2C(
        stage=UNKNOWN_STAGE, cmd=amaas.grpc.scan_pb2.CMD_QUIT
    )

    pipeline.set_message(server_resp)
    with pytest.raises(AMaasException) as exc_info:
        next(amaas.grpc._generate_messages(pipeline, f, True, stats), None)
    assert exc_info.value.args[0] == AMaasErrorCode.MSG_ID_ERR_UNKNOWN_STAGE
    assert exc_info.value.args[1] == UNKNOWN_STAGE


#
# Testing the SDK scan_file method sucessfully scans a file with no virus.
#
def test_scan_file_success():
    dir_path = os.path.dirname(os.path.realpath(__file__))
    handle = grpc.insecure_channel(f"localhost:{SERVER_PORT}")
    response = amaas.grpc.scan_file(handle, dir_path + "/fake_server_cert.pem")
    assert json.loads(response)["scanResult"] == 0


#
# Testing the SDK scan_file method sucessfully failed with MSG_ID_ERR_FILE_NOT_FOUND exception.
#
def test_scan_file_not_found():
    handle = grpc.insecure_channel(f"localhost:{SERVER_PORT}")
    NOT_EXIST_FILE = f"{str(uuid.uuid4())}.txt"
    with pytest.raises(AMaasException) as exc_info:
        amaas.grpc.scan_file(handle, NOT_EXIST_FILE)
    assert exc_info.value.args[0] == AMaasErrorCode.MSG_ID_ERR_FILE_NOT_FOUND
    assert exc_info.value.args[1] == NOT_EXIST_FILE


#
# Testing the SDK scan_file method failed with MSG_ID_ERR_FILE_NO_PERMISSION exception
#
def test_scan_file_no_permission():
    handle = grpc.insecure_channel(f"localhost:{SERVER_PORT}")
    NOT_PERMISSION_FILE = tempfile.NamedTemporaryFile().name
    open(NOT_PERMISSION_FILE, "x")
    os.chmod(NOT_PERMISSION_FILE, 0x000)
    with pytest.raises(AMaasException) as exc_info:
        amaas.grpc.scan_file(handle, NOT_PERMISSION_FILE)
    assert exc_info.value.args[0] == AMaasErrorCode.MSG_ID_ERR_FILE_NO_PERMISSION
    assert exc_info.value.args[1] == NOT_PERMISSION_FILE
    os.remove(NOT_PERMISSION_FILE)


#
# Testing the SDK scan_buffer method sucessfully scans a file with no virus.
#
def test_scan_buffer_success():
    handle = grpc.insecure_channel(f"localhost:{SERVER_PORT}")
    with open(TEST_DATA_FILE_NAME, mode="rb") as bfile:
        buffer = bfile.read()
        response = amaas.grpc.scan_buffer(handle, buffer, TEST_DATA_FILE_NAME)
        assert json.loads(response)["scanResult"] == 0


#
# Testing the SDK scan_buffer method sucessfully scans a file with virus.
#
def test_scan_buffer_virus():
    handle = grpc.insecure_channel(f"localhost:{SERVER_PORT}")
    with open(TEST_DATA_FILE_NAME, mode="rb") as bfile:
        buffer = bfile.read()
        response = amaas.grpc.scan_buffer(
            handle, buffer, MockScanServicer.IDENTIFIER_VIRUS
        )
        jobj = json.loads(response)
        assert jobj["scanResult"] == 1
        assert jobj["foundMalwares"] == ["virus1", "virus2"]


#
# Test different exceptions throwed by scan_buffer API.
#
@pytest.mark.parametrize(
    "error_type, expected_exception",
    [
        (
            MockScanServicer.IDENTIFIER_EXCEED_RATE,
            [AMaasErrorCode.MSG_ID_ERR_RATE_LIMIT_EXCEEDED],
        ),
        (
            MockScanServicer.IDENTIFIER_UNKNOWN_CMD,
            [AMaasErrorCode.MSG_ID_ERR_UNKNOWN_CMD, MockScanServicer.UNKNOWN_CMD],
        ),
        (
            MockScanServicer.IDENTIFIER_GRPC_ERROR,
            [AMaasErrorCode.MSG_ID_GRPC_ERROR, grpc.StatusCode.INTERNAL.value[0]],
        ),
    ],
)
def test_scan_buffer_exceptions(error_type, expected_exception):
    handle = grpc.insecure_channel(f"localhost:{SERVER_PORT}")
    with open(TEST_DATA_FILE_NAME, mode="rb") as bfile:
        buffer = bfile.read()
        with pytest.raises(AMaasException) as exc_info:
            amaas.grpc.scan_buffer(handle, buffer, error_type)
        for cnt in range(len(expected_exception)):
            assert exc_info.value.args[cnt] == expected_exception[cnt]
